# -*- coding: utf-8 -*-

import pygame
from pygame.sprite import Sprite

from . import actors, SCREEN_WIDTH, SCREEN_HEIGHT, WAIT, EXPLORE, HUNT, RUN, LEWD
from vec2d import vec2d
from copy import copy

class Actor(Sprite):
    def __init__(self, screen, init_position,
            init_direction, behavior, dna, debug = False):
        """ Create a new Actor.
        
            screen: 
                The screen on which the actor lives (must be a 
                pygame Surface object, such as pygame.display).
            
            init_position:
                A vec2d or a pair specifying the initial position
                of the actor on the screen.
            
            init_direction:
                A vec2d or a pair specifying the initial direction
                of the actor. Must have an angle that is a 
                multiple of 45 degres.
                
            behavior:
                An object determining the runtime-constant behavior.
                
            dna:
                An object determining the runtime-variable behavior.
                
            debug:
                Switches debug mode on or off.
        """
        Sprite.__init__(self)

        self.screen = screen
        self.pos = vec2d(init_position)
        self.direction = vec2d(init_direction).normalized()
        self.behavior = behavior
        self.behavior.actor = self
        self.dna = dna
        self.debug = debug
        self.relative_distance = 0 #needed for faster algorithm

        if self.debug:
            print init_direction, self.direction

        # base_image holds the original image, positioned to angle 0.
        # image will be rotated. 
        self.base_image = pygame.image.load(self.behavior.image_filename).convert_alpha()
        self.image = self.base_image
        self.image_width, self.image_height = self.image.get_size()

        SCALE = 1.0

        self.rect = pygame.Rect(self.pos.x % SCREEN_WIDTH - 0.5 * SCALE * self.image_width,
                                 self.pos.y % SCREEN_HEIGHT - 0.5 * SCALE * self.image_height,
                                 SCALE * self.image_width,
                                 SCALE * self.image_height)

        self.state = EXPLORE
        self.target = None
        self.energy = self.dna.energy

    def update(self, time_passed, x_list, y_list):
        """ Update the actor.
        
            time_passed:
                The time passed (in ms) since the previous update
        """

        if self.state == WAIT:
            self.energy -= 0.1 * self.dna.energyneed * time_passed
        elif self.state == EXPLORE or self.state == LEWD:
            self.energy -= 0.5 * self.dna.energyneed * abs(self.dna.speed + self.dna.boost) * self.dna.speed * time_passed
        elif self.state == RUN or self.state == HUNT:
            self.energy -= 1 * self.dna.energyneed * abs(self.dna.speed + self.dna.boost) * self.dna.speed * time_passed

        if self.energy <= 0:
            actors.remove(self)
            return

        self.select_target(x_list, y_list)

        # Maybe it's time to change the direction?
        self.behavior.change_direction(time_passed)

        self.munch(x_list, y_list)

        # Make the creep point in the correct direction.
        # Since our direction vector is in screen coordinates 
        # (i.e. right bottom is 1, 1), and rotate() rotates 
        # counter-clockwise, the angle must be inverted to 
        # work correctly.

        self.image = pygame.transform.rotate(
            self.base_image, -self.direction.angle)

        # When the image is rotated, its size is changed.
        # We must take the size into account for detecting 
        # collisions with the walls.

        self.image_width, self.image_height = self.image.get_size()

        # Compute and apply the displacement to the position 
        # vector. The displacement is a vector, having the angle
        # of self.direction (which is normalized to not affect
        # the magnitude of the displacement)

        if self.state == RUN or self.state == HUNT or self.state == LEWD:
            displacement = vec2d(
                self.direction.x * abs(self.dna.speed + self.dna.boost) * time_passed,
                self.direction.y * abs(self.dna.speed + self.dna.boost) * time_passed)
            self.pos += displacement

        elif self.state == EXPLORE:
            displacement = vec2d(
                self.direction.x * abs(self.dna.speed) * time_passed,
                self.direction.y * abs(self.dna.speed) * time_passed)
            self.pos += displacement



        if self.debug:
            print self.direction, self.speed, time_passed

        SCALE = 1.5
        self.rect = pygame.Rect(self.pos.x % SCREEN_WIDTH - 0.5 * SCALE * self.image_width,
                                 self.pos.y % SCREEN_HEIGHT - 0.5 * SCALE * self.image_height,
                                 SCALE * self.image_width, SCALE * self.image_height)

        bounds_rect = self.screen.get_rect().inflate(-self.image_width, -self.image_height)

        if self.pos.x < bounds_rect.left:
            self.pos.x = bounds_rect.left
            self.direction.x *= -1
            self.boost = 0.0
        elif self.pos.x > bounds_rect.right:
            self.pos.x = bounds_rect.right
            self.direction.x *= -1
            self.boost = 0.0
        elif self.pos.y < bounds_rect.top:
            self.pos.y = bounds_rect.top
            self.direction.y *= -1
            self.boost = 0.0
        elif self.pos.y > bounds_rect.bottom:
            self.pos.y = bounds_rect.bottom
            self.direction.y *= -1
            self.boost = 0.0


        self.blitme()

    def spawn(self, other):
        i = 1.
        while self.energy > self.dna.energy * 1.5 * i and other.energy > other.dna.energy * 1.5 * i:

            dna = self.dna.cross(other.dna)
            actors.add(Actor(self.screen, self.pos, self.direction, copy(self.behavior), dna))
            self.energy -= (dna.energy * i) / 2
            other.energy -= (dna.energy * i) / 2
            i += 0.25

    def select_target(self, x_list, y_list):
        neighbors = self.behavior.look_around(self.dna.sight_radius, x_list, y_list)

        hunter = None
        food = None
        friends = []

        if neighbors != None:
            for actor in neighbors:
                if actor.behavior.type == self.behavior.type:
                    if len(friends) < 6:
                        friends.append(actor)
                if food == None and actor.behavior.type in self.behavior.food:
                    food = actor
                if hunter == None and actor.behavior.type in self.behavior.hunters:
                    hunter = actor
        else:
            self.state = EXPLORE
            self.target = None

        if hunter != None:
            if food != None:
                part_run = self.dna.sight_radius / (hunter.pos - self.pos).get_length() * self.dna.weight_run
                part_eat = self.dna.sight_radius / (hunter.pos - self.pos).get_length() * self.dna.weight_food
                if part_run > part_eat:
                    self.state = RUN
                    self.target = hunter
                else:
                    self.state = HUNT
                    self.target = food
            else:
                self.state = RUN
                self.target = hunter
        else:
            if food != None:
                self.state = HUNT
                self.target = food
            else:
                self.state = EXPLORE
                self.target = friends

    def munch(self, x_list, y_list):
        food = self.behavior.look_around(self.dna.munch_radius, x_list, y_list)
        if food != None:
            for actor in food:
                if actor.behavior.type in self.behavior.food:
                    self.energy += actor.energy
                    actors.remove(actor)
                if actor.behavior.type == self.behavior.type:
                    self.spawn(actor)

    def blitme(self):
        """ Blit the actor onto the screen that was provided in
            the constructor
        """

        # The actor image is placed at self.pos.
        # To allow for smooth movement even when the actor rotates
        # and the image size changes, its placement is always
        # centered.

        draw_pos = self.image.get_rect().move(
            self.pos.x % SCREEN_WIDTH - self.image_width / 2,
            self.pos.y % SCREEN_HEIGHT - self.image_height / 2)
        self.screen.blit(self.image, draw_pos)
